'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = lodash;

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _config = require('./config');

var _config2 = _interopRequireDefault(_config);

var _importModule = require('./importModule');

var _importModule2 = _interopRequireDefault(_importModule);

var _mapping = require('./mapping');

var _mapping2 = _interopRequireDefault(_mapping);

var _Store = require('./Store');

var _Store2 = _interopRequireDefault(_Store);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/** The error message used when chain sequences are detected. */
var CHAIN_ERROR = ['Lodash chain sequences are not supported by babel-plugin-lodash.', 'Consider substituting chain sequences with composition patterns.', 'See https://medium.com/making-internets/why-using-chain-is-a-mistake-9bc1f80d51ba'].join('\n');

/*----------------------------------------------------------------------------*/

function lodash(_ref) {
  var types = _ref.types;


  var identifiers = {
    'PLACEHOLDER': types.identifier('placeholder'),
    'UNDEFINED': types.identifier('undefined')
  };

  /**
   * Used to track variables built during the AST pass. We instantiate these in
   * the `Program` visitor in order to support running the plugin in watch mode
   * or on multiple files.
   *
   * @type Store
   */
  var store = new _Store2.default();

  function getCallee(_ref2) {
    var parentPath = _ref2.parentPath;

    // Trace curried calls to their origin, e.g. `fp.partial(func)([fp, 2])(1)`.
    while (!parentPath.isStatement()) {
      if (parentPath.isCallExpression()) {
        var result = parentPath.node.callee;
        while (types.isCallExpression(result)) {
          result = result.callee;
        }
        return result;
      }
      parentPath = parentPath.parentPath;
    }
  }

  /*--------------------------------------------------------------------------*/

  function importer(_ref3) {
    var defaults = _ref3.defaults;
    var members = _ref3.members;

    _lodash2.default.each(members, function (_ref4) {
      var pkgStore = _ref4.pkgStore;
      var specPath = _ref4.specPath;

      // Import module.
      var node = specPath.node;

      var name = node.local.name;
      var binding = specPath.scope.getBinding(name);
      var identifier = (0, _importModule2.default)(pkgStore, node.imported.name, specPath.hub.file);
      var isLodash = pkgStore.isLodash();

      // Replace the old member with its new identifier.
      _lodash2.default.each(binding.referencePaths, function (refPath) {
        if (isLodash && name == 'chain' && refPath.parentPath.isCallExpression()) {
          throw refPath.buildCodeFrameError(CHAIN_ERROR);
        }
        refPath.replaceWith(types.clone(identifier));
      });
    });

    _lodash2.default.each(defaults, function (_ref5) {
      var pkgStore = _ref5.pkgStore;
      var specPath = _ref5.specPath;

      var name = specPath.node.local.name;
      var binding = specPath.scope.getBinding(name);
      var isLodash = pkgStore.isLodash();

      _lodash2.default.each(binding.referencePaths, function (refPath) {
        var parentPath = refPath.parentPath;


        if (parentPath.isMemberExpression()) {
          var key = refPath.parent.property.name;
          if (isLodash && key == 'chain' && parentPath.parentPath.isCallExpression()) {
            throw refPath.buildCodeFrameError(CHAIN_ERROR);
          }
          // Import and replace the old member with its new identifier.
          var identifier = (0, _importModule2.default)(pkgStore, key, refPath.hub.file);
          parentPath.replaceWith(types.clone(identifier));
        } else if (isLodash) {
          var callee = getCallee(refPath);
          if (callee && callee.name == name) {
            throw refPath.buildCodeFrameError(CHAIN_ERROR);
          }
          refPath.replaceWith(callee ? types.memberExpression(callee, identifiers.PLACEHOLDER) : identifiers.UNDEFINED);
        }
      });
    });
  }

  var importVisitor = {
    ImportDeclaration: function ImportDeclaration(path) {
      var _this = this;

      var node = path.node;

      var pkgStore = store.get(node.source.value);

      if (!pkgStore) {
        return;
      }
      _lodash2.default.each(path.get('specifiers'), function (specPath) {
        var key = specPath.isImportSpecifier() ? 'members' : 'defaults';
        _this[key].push({ pkgStore: pkgStore, specPath: specPath });
      });
      // Remove old import.
      path.remove();
    }
  };

  var visitor = {
    Program: function Program(path, state) {
      var _$assign = _lodash2.default.assign(_mapping2.default, (0, _config2.default)(state.opts));

      var ids = _$assign.ids;

      if (_lodash2.default.isEmpty(ids)) {
        throw new Error('Cannot find module');
      }
      // Clear tracked Lodash method imports and variables.
      _importModule2.default.cache.clear();
      store.clear();

      // Populate module paths per package.
      _lodash2.default.each(ids, function (id) {
        store.set(id);
        _mapping2.default.modules.get(id).forEach(function (value, key) {
          store.set(id + '/' + key);
        });
      });

      var imports = {
        'declarations': [],
        'defaults': [],
        'members': []
      };

      path.traverse(importVisitor, imports);
      importer(imports);
    },
    ExportNamedDeclaration: function ExportNamedDeclaration(path) {
      var node = path.node;

      var pkgPath = _lodash2.default.get(node, 'source.value');
      var pkgStore = store.get(pkgPath);

      if (!pkgStore) {
        return;
      }
      node.source = null;
      _lodash2.default.each(node.specifiers, function (spec) {
        spec.local = (0, _importModule2.default)(pkgStore, spec.local.name, path.hub.file);
      });
    }
  };

  return { visitor: visitor };
};
module.exports = exports['default'];